#
#   Imandra Inc.
#
#   main.py - Main entrypoint for CodeLogician
#

import asyncio
import json
import logging
import os
import shutil
from pathlib import Path
from typing import Annotated, Final

import typer
from rich import print as printr

from codelogician.commands.cmd_doc import app as doc_app
from codelogician.commands.cmd_eval import app as eval_app
from codelogician.commands.cmd_metamodel import app as metamodel_app
from codelogician.commands.cmd_model import app as model_app
from codelogician.commands.cmd_sketches import app as sketches_app
from codelogician.commands.cmd_strategy import app as strategy_app

# logging.basicConfig(level=logging.INFO)
logging.getLogger('httpx').setLevel(logging.WARNING)
logging.getLogger('httpcore').setLevel(logging.WARNING)
logging.getLogger('fsevents').setLevel(logging.WARNING)
# logging.getLogger("strategy.pyiml_strategy").setLevel(logging.WARNING)
# logging.getLogger("strategy.cl_worker").setLevel(logging.WARNING)
# logging.getLogger("strategy.metamodel").setLevel(logging.WARNING)
log = logging.getLogger(__name__)


def setup_logging(debug: bool = False):
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.WARNING,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        handlers=[logging.FileHandler('cl.log'), logging.StreamHandler()],
    )


CL_HELP: Final = """
[bold deep_sky_blue1][italic]CodeLogician[/italic] is the neurosymbolic agentic governance framework for AI-powered coding.[/bold deep_sky_blue1] :rocket:

[white]***[bold italic purple]It helps your coding agent think logically about the code it's producing and test cases it's generating.[/bold italic purple]***
The fundamental flaw that all LLM-powered assistants have is the reasoning they're capable of is based on statistics,
while you need rigorous logic-based automated reasoning.

- Generated code is based on explainable logic, not pure statistics
- Generated test cases are generated come with quantitative coverage metrics
- Generated code is consistent with the best security practices leveraging formal verification

To run [bold italic deep_sky_blue1]CodeLogician[/bold italic deep_sky_blue1], please obtain an [bold italic deep_sky_blue1]Imandra Universe API[/bold italic deep_sky_blue1] key available
(there's a free starting plan) at [bold italic deep_sky_blue1]https://universe.imandra.ai[/bold italic deep_sky_blue1] and make
sure it's available in your environment as [bold italic deep_sky_blue1]`IMANDRA_UNIVERSE_KEY`[/bold italic deep_sky_blue1].

[bold deep_sky_blue1 italic]Three typical workflows[/bold deep_sky_blue1 italic]:
1. [bold deep_sky_blue1 italic]DIY mode[/bold deep_sky_blue1 italic] - this is where your agent (e.g. Grok) uses the CLI to:
  - Learn how to use IML/ImandraX via `doc` command (e.g. `codelogician doc --help`)
  - Synthesizes IML code and uses the `eval` command to evaluate it
  - If there're errors, use `codelogician doc view errors` command to study how to correct the errors and re-evalute the code

2. [bold deep_sky_blue1 italic]Agent/multi-agent mode[/bold deep_sky_blue1 italic] - CodeLogician Agent is a Langgraph-based agent for automatically formalizing source code.
  - With `agent` command you can formalize a single source code file (e.g. `codelogician agent PATH_TO_FILE`)
  - With `multiagent` command you can formalize a whole directory (e.g. `codelogician agent PATH_TO_DIR`)

3. [bold deep_sky_blue1 italic]Server[/bold deep_sky_blue1 italic] - this is a "live" and interactive version of the `multiagent` command, but one you can interact with and one that
"listens" to live updates and automatically updates formalization as necessary. You can start the server and connect to it
with the TUI (we recommend separate terminal screens).

Learn more at[/white] [bold italic deep_sky_blue1]https://www.codelogician.dev![/bold italic deep_sky_blue1]
"""

app = typer.Typer(name='CodeLogician', help=CL_HELP, rich_markup_mode='rich')

# fmt: off
app.add_typer(model_app    , name='model'    , help='Model-related commands (e.g. view and list)')
app.add_typer(strategy_app , name='strategy' , help='Server function for creating and managing strategies')
app.add_typer(metamodel_app, name='metamodel', help='Server metamodel-related functions')
app.add_typer(sketches_app , name='sketches' , help='Server API for creating and managing sketches')
app.add_typer(doc_app      , name='doc'      , help='Documentation, guides and ImandraX reference')
app.add_typer(eval_app     , name='eval'     , help='Evaluate IML file via ImandraX API')
# fmt: on


# @app.callback(invoke_without_command=True)
# def default(ctx: typer.Context):
#  print (ctx.command.params)
#  typer.echo(ctx.get_help())


@app.command(
    name='sample',
    help='Creates a sample Python project in specified directory (relative to the current directory)',
)
def run_sample_cmd(
    tgt_dir: Annotated[
        str, typer.Argument(help='Target directory where to place the sample project')
    ],
):
    """
    Create a sample project
    """

    full_tgt_dir = Path(os.path.join(os.getcwd(), tgt_dir)).resolve()

    if os.path.exists(full_tgt_dir):
        typer.secho('⚠️ Sample directory already exists!')
    else:
        os.makedirs(full_tgt_dir)

    if not os.path.exists(full_tgt_dir):
        print(f'🛑 Failed to create sample project directory: {full_tgt_dir}')
        return

    source_dir = 'data/sample_bank_app'

    for filepath in os.listdir(source_dir):
        if filepath.endswith('.py'):
            tgt_filepath = os.path.join(full_tgt_dir, Path(filepath).name)
            try:
                shutil.copyfile(os.path.join(source_dir, filepath), tgt_filepath)
            except Exception as e:
                typer.secho(
                    f'🛑 Failed to copy project file: {filepath} {tgt_filepath}: {e}',
                    err=True,
                )
                raise typer.Exit(1)

    typer.secho(f'✅ Created sample project in: {full_tgt_dir}')


# fmt: off
@app.command(name='server', help='Start the server')
def run_server_cmd(
    dir   : Annotated[str, typer.Argument(help='Starting directory')],
    state : Annotated[str | None, typer.Option(help='Server state path')] = None,
    clean : Annotated[bool, typer.Option(help='Disregard existing strategy caches')] = False,
    config: Annotated[str, typer.Option(help='Configuration path')] = 'config/server_config.yaml',
    debug : Annotated[bool, typer.Option(help='Debug mode')] = False,
    addr  : Annotated[str, typer.Option(help='Server host/port')] = 'http://127.0.0.1:8000',
):
    from codelogician.server.main import run_server

    setup_logging(debug=debug)
    run_server(dir=dir, state=state, clean=clean, config=config, addr=addr)
# fmt: on


# fmt: off
@app.command(
    name='multiagent',
    help='Run the CL server in autoformalization mode for a directory and then quit.',
)
def run_oneshot_cmd(
    dir   : Annotated[str, typer.Argument(help='Target directory')],
    clean : Annotated[bool, typer.Option(help='Start clean by disregarding any existing cache files')] = False,
    config: Annotated[str, typer.Option(help='Server configuration YAML file')] = 'config/server_config.yaml',
    debug : Annotated[bool, typer.Option(help='Debug mode')] = False,
):
    from codelogician.server.oneshot import run_oneshot

    setup_logging(debug=debug)
    run_oneshot(dir, clean, config)
# fmt: on


@app.command(name='tui', help='Run the TUI')
def run_tui_cmd(
    addr: Annotated[
        str, typer.Option(help='Server host/port')
    ] = 'http://127.0.0.1:8000',
):
    """
    Run the TUI
    """

    from codelogician.ui.main import run_tui

    run_tui(addr)


@app.command(name='agent', help='Formalize a file via CodeLogician agent')
def run_frmfile_cmd(
    file: Annotated[
        Path, typer.Argument(help='Path to a source code file to run the CL agent on')
    ],
    outfile: Annotated[
        Path,
        typer.Argument(help='Output file path to save the resulting CL Agent state'),
    ] = Path('cl_result.json'),
    artifacts: Annotated[
        bool,
        typer.Option(
            help='Should we generate artifacts like Verification Goals and Decomposition requests'
        ),
    ] = True,
):
    """
    Read the provided file and run CodeLogician Agent on it, saving the result in `outfile` file.
    """

    from imandra.u.agents.code_logician import SUPPORTED_LANGUAGES

    from codelogician.tools.cl_caller import call_cl_agent

    if not file.exists():
        typer.secho(f'🛑 File does not exist: {file}')
        raise typer.Exit(1)

    language = None
    for lang, ext in SUPPORTED_LANGUAGES:
        if file.suffix == ext:
            language = lang
            break

    if language is None:
        typer.secho(
            f'🛑 Unsupported language with extension: {file.suffix}. Only the following are supported: {SUPPORTED_LANGUAGES}'
        )
        raise typer.Exit(1)

    try:
        src_code = file.read_text()
    except Exception as e:
        typer.secho(f'🛑 Failed to read the file: {e}')
        raise typer.Exit(1)

    try:
        res = asyncio.run(
            call_cl_agent(src_code=src_code, language=language, artifacts=artifacts)
        )
    except Exception as e:
        typer.secho(
            f'🛑 Caught exception during the call: {e}', fg=typer.colors.RED, err=True
        )
        raise typer.Exit(1)

    if res is None:
        typer.secho('🛑 Failed to execute the CL agent. Received `None`', err=True)
        raise typer.Exit(1)

    printr(res)

    try:
        with open(outfile, 'w') as outfile_file:
            data = res.model_dump(mode='json')
            fields = [
                'status',
                'src_code',
                'src_lang',
                'eval_res',
                'vgs',
                'region_decomps',
                'iml_code',
                'iml_model',
            ]
            select_data = {k: data[k] for k in fields if k in data}
            print(json.dumps(select_data, indent=4), file=outfile_file)
            typer.secho(f'✅ Written results to {outfile}!')

    except Exception as e:
        typer.secho(f'🛑 Failed to print to out file: {e}', err=True)
        raise typer.Exit(1)


def run_codelogician():
    app()


if __name__ == '__main__':
    app()



#
#   Imandra Inc.
#
#   search.py
#

import rich.repr
from pydantic import BaseModel


@rich.repr.auto
class SearchResult(BaseModel):
    """
    Search result returned from a strategy
    """

    distance: float
    text: str
    rel_path: str
    match_entity_type: str
    strat_name: str



#
#   Imandra Inc.
#
#   meta_model.py
#

import json
import logging
from collections import defaultdict
from functools import cmp_to_key
from pathlib import Path

from imandra.u.agents.code_logician.base import FormalizationStatus
from pydantic import BaseModel, model_validator
from rich.console import Group
from rich.panel import Panel
from rich.pretty import Pretty
from rich.text import Text

from ..dep_tools.python import build_py_dep_graph
from ..server.events import FileSystemEvent, FileSystemEventType
from ..tools.filesystem import FileSystem
from ..util import filter, fst, intersperse, snd
from .cl_agent_state import CLAgentState, CLResult
from .model import Model, ModelList
from .model_task import ModelTask
from .sketch import Sketch

log = logging.getLogger(__name__)


class MetaModel(BaseModel):
    """
    Container for the set of models representing the whole directory
    """

    src_dir_abs_path: str
    models: dict[str, Model]
    language: str = 'Python'

    @model_validator(mode='after')
    def link_model_deps(self) -> 'MetaModel':
        """
        When store states to disk, we store relative paths for
        dependencies/rev_dependencies. Here we substitute them with actual model objects.
        """

        for model in self.models.values():
            model_deps, model_rev_deps = [], []

            for dep in model.dependencies:
                model_deps.append(dep)

            for dep in model.rev_dependencies:
                model_rev_deps.append(dep)

            model.dependencies = model_deps
            model.rev_dependencies = model_rev_deps

        return self

    def search(
        self, query_vector: list[float], num_top_hits: int = 5
    ) -> list[tuple[Model, tuple[float, str]]]:
        """
        Return the list of models more whose embedding vectors are closest to the query, along with
        the distance.
        """

        # TODO This right now just gets us the top hit from each model
        # but, we should instead aggregate them across models and get
        # the top hits that way
        # fmt: off
        def snd_not_none(xy): return snd(xy) is not None
        def with_dist(m): return m, m.calc_embedding_distance(query_vector)
        def dist_of_tuple(t): return fst(snd(t))
        # fmt: on
        # build and sort list of tuples, then select top_k
        tuples = filter(snd_not_none, map(with_dist, self.models.values()))
        return sorted(tuples, key=dist_of_tuple)[:num_top_hits]

    def get_paths_with_dirs(self):
        """
        Return a list of paths with directory information
        """

        FILE_MARKER = '<files>'

        def attach(branch, trunk, model):
            parts = branch.split('/', 1)
            if len(parts) == 1:  # branch is a file
                trunk[FILE_MARKER].append(model)
            else:
                node, others = parts
                if node not in trunk:
                    trunk[node] = defaultdict(dict, ((FILE_MARKER, []),))
                attach(others, trunk[node], model)

        main_dict = defaultdict(dict, ((FILE_MARKER, []),))

        for path in sorted(self.models.keys()):
            attach(path, main_dict, self.models[path])

        return main_dict

    def list_models(self) -> ModelList:
        """
        Return the list of all models
        """
        return ModelList(models=list(self.models.values()))

    def vgs(self):
        """
        Return all verification goals for all the models
        """
        return {m.rel_path: m.verification_goals() for m in self.models.values()}

    def decomps(self):
        """
        Return decomposition requests for all the models
        """
        return {m.rel_path: m.decomps() for m in self.models.values()}

    def opaques(self):
        """
        Return all the opaque functions for all the models
        """
        return {m.rel_path: m.opaque_funcs() for m in self.models.values()}

    def apply_cl_result(self, cl_result: CLResult) -> None:
        """
        Applies the result of running CodeLogician on a model
        """

        if cl_result.task.rel_path in self.models:
            model = self.models[cl_result.task.rel_path]

            # Rememeber that we may have multiple tasks outstanding simultaneously
            # so, we always keep the last one that we use
            if model.outstanding_task_ID == cl_result.task.task_id:
                dependencies = cl_result.task.dependencies
                model.apply_agent_state(cl_result.agent_state, dependencies)
                model.outstanding_task_ID = None

                log.info('CLResult applied to model successfully!')
            else:
                log.warning(
                    f'Model with path={cl_result.task.rel_path} has a newer Task ID, so ignoring this one!'
                )
        else:
            log.error(
                f'Attempting to apply CL Result to a non-existent model path: {cl_result.task.rel_path}!'
            )

    def update_model(self, rel_path: str, state: CLAgentState) -> None:
        """
        Update the model with a new agent state (formalization state form CL agent)
        """

        if rel_path in self.models:
            self.models[rel_path].apply_agent_state(state)
            log.info(
                f'AgentState successfully updated for model with path=[{rel_path}]'
            )
        else:
            log.error(
                f'Attempting to set agent state to a non-existent model path {rel_path}'
            )

    def get_embedding_tasks(self) -> list[ModelTask]:
        """
        Generate list of CL tasks to generate embeddings
        """

        tasks = []
        for model in self.models.values():
            if model.needs_embedding_update():
                tasks.append(model.gen_embedding_task())

        return tasks

    def get_next_tasks(self, user_wait_time: int | None = None) -> list[ModelTask]:
        """
        Create a list of formalization tasks for the strategy to send to the CL agent.

        We're looking for models that need:
        - Have a
        - Formalization done
        - Have no formalization tasks outstanding (`model.outstanding_task_ID == None`)
        - Have no dependencies that need formalization

        - user_wait_time - how much we should wait since the last src code/iml model update -
        this is used to determine whether `src_code` and/or 'iml_code' have changed.
        """

        log.debug('About to `get_next_tasks`')

        tasks: list[ModelTask] = []

        for model in self.models.values():
            if (
                model.user_iml_change_ready(user_wait_time)
                and model.outstanding_task_ID is None
            ):
                log.debug(f'Model {model.rel_path} will need a user IML override task')
                model_task = model.gen_iml_user_update_task()
                if model_task:
                    tasks.append(model_task)
                    log.info(
                        f'Added model task for manually setting IML code [{model.rel_path}]'
                    )
                else:
                    log.error(
                        f"We should've gotten a ModelTask user IML change for this model, but didnt! [{model.rel_path}]"
                    )
                    continue

                log.debug(
                    f'The model [{model.rel_path}] has unprocessed user set IML code.'
                )

            else:
                # Let's now consider the "regular" CL task
                log.debug(
                    f'Will now check if model {model.rel_path} needs formalization'
                )
                needs_formalization = bool(model.formalization_reasons(user_wait_time))
                deps_dont_need_formalization = not (model.deps_need_formalization())

                if (
                    needs_formalization
                    and deps_dont_need_formalization
                    and model.outstanding_task_ID is None
                ):
                    log.info(
                        f'{model.rel_path} needs formalization: [{model.formalization_reasons(user_wait_time)}]'
                    )
                    tasks.append(model.gen_formalization_task())

                    log.debug(f'Model [{model.rel_path}] has a formalization task!')
                else:
                    log.debug('Nope')

        log.debug(f'Returning {len(tasks)} formalization tasks!')

        return tasks

    def calc_upstream_affected(self, rel_path: str) -> int:
        """
        Return the number of models affected by the model with specified 'rel_path'
        """

        if rel_path not in self.models:
            raise Exception(f'Specified model [{rel_path}] is not present!')

        def helper(current_path: str, visited: list[str]):
            """Our helper function to help us traverse the models"""

            for path in self.models:
                if path in visited:
                    continue

                found = False

                for d in self.models[path].dependencies:
                    if current_path == d.rel_path:
                        found = True
                        break

                if found:
                    visited.append(path)
                    visited = helper(path, visited)

            return visited

        return len(helper(rel_path, []))

    def gen_listing(self, category: str = 'frm_status'):
        """
        Return a sorted list of models by a specified category.

        Possible values are:
        - Formalization Status [value='frm_status'] - formalization status
        - Upstream dependencies [value='upstream'] - number of upstream dependencies
        - Number of opaque functions [value='opaques'] - number of opaque functions
        - Number of failed VGs [value='failed_vgs'] - number of failed verification goals
        """

        supported = ['frm_status', 'upstream', 'opaques', 'failed_vgs']

        # Let's make sure the category is supported first
        if category.lower() not in supported:
            raise Exception(
                f'Supplied category is not supported: {category}. Possible values are: {supported}'
            )

        stats = {}

        # This should get us the number of models that are ultimatey affected by this model
        for rel_path in self.models:
            upstream = self.calc_upstream_affected(rel_path)
            stats[rel_path] = {'path': rel_path, 'upstream': upstream}

        for rel_path in self.models:
            modelStats = self.models[rel_path].gen_stats()

            # Let's now append the statistics here
            for key in modelStats:
                stats[rel_path][key] = modelStats[key]

        def compare(s1, s2):
            try:
                if category.lower() == 'frm_status':
                    return s1['frm_status'] >= s2['frm_status']
                elif category.lower() == 'upstream':
                    return s1['upstream'] >= s2['upstream']
                elif category.lower() == 'opaques':
                    return s1['num_opaques'] >= s2['num_opaques']
                else:
                    return s1['failed_vgs'] >= s2['failed_vgs']
            except Exception as e:
                raise Exception(f'Failed to compare model statistcs: {str(e)}')

        return list(sorted(stats.values(), key=cmp_to_key(compare)))

    def abs_path_in_models(self, path: str) -> bool:
        """
        Return True if the specified path in the models
        """
        return self.make_path_relative(path) in self.models

    def make_path_relative(self, path: str) -> str:
        """
        Return True/False if the specified path is relative to the base source code directory
        """
        return str(Path(path).relative_to(self.src_dir_abs_path))

    def get_model_by_path(self, rel_path: str) -> Model | None:
        """
        Returns Model object for specified relative path, None if it's missing
        """

        if rel_path in self.models:
            return self.models[rel_path]
        return None

    def gather_deep_dependencies(self, rel_path: str) -> list[Model]:
        """
        Return list of all "deep" dependencies for the specified relative path
        """

        if rel_path not in self.models:
            return []

        def rec_helper(curr_model: Model):
            res = []
            for d in curr_model.dependencies:
                res.extend(rec_helper(d) + [d])
            return res

        curr_model = self.models[rel_path]
        return list(set(rec_helper(curr_model) + [curr_model]))

    def gen_consolidated_model(self, rel_path: str) -> tuple[str, dict[str, Model]]:
        """
        Generate a single consolidated model from the achnor model specified by (rel_path).
        """

        if rel_path not in self.models:
            raise Exception(f'Specified model {rel_path} is not a model!')

        # we'll gather the relevant models (subgraph) and then sort it
        relevant_models: list[Model] = self.gather_deep_dependencies(rel_path)
        sorted_list: list[Model] = Model.do_topological_sort(relevant_models)

        # now we'll consolidate the code
        iml = ''
        for model in sorted_list:
            iml += (
                f'(* Starting {model.rel_path} *)\n'
                + (str(model.iml_code()) if model.iml_code() else '')
                + f'\n(* Ending {model.rel_path} *)\n'
            )

        return iml, {m.rel_path: m for m in relevant_models}

    def process_filesystem_event(self, event: FileSystemEvent):
        """
        Note that the provided paths are absolute!
        We receive events that concern both - Python and IML files.
        """

        # Let's figure out if we have a source code file or we have
        #

        ext = Path(event.abs_path1).suffix
        if ext == '.py':
            self.process_fs_event_src_code(event)
        elif ext == '.iml':
            self.process_fs_event_iml_code(event)
        else:
            log.warning(
                f'MetaModel asked to process FS event for unknown extension: {ext}'
            )

    def process_fs_event_iml_code(self, event: FileSystemEvent):
        """
        Process FileSystem event when it's for a model
        """

        if event.action_type in [
            FileSystemEventType.CREATED,
            FileSystemEventType.MODIFIED,
        ]:
            # Let's first figure out what the Model path would be for this guy

            rel_path = self.make_path_relative(event.abs_path1)
            model_path = str(Path(rel_path).with_suffix('.py'))

            if model_path in self.models:
                try:
                    with open(event.abs_path1) as infile:
                        imlCode = infile.read()
                except Exception as e:
                    log.error(f'Failed to read IML file: {event.abs_path1}: {e}')
                    return

                self.models[model_path].set_iml_model(imlCode, record_time=True)
                self.models[model_path].expected_iml_on_disk = imlCode

                log.info(f'Updated IML code for {model_path}')
            else:
                log.info(
                    f'Received IML file for unknown model. Implied model path is: {model_path}'
                )
        else:
            log.info(
                f'We only support CREATED/MODIFIED FS events for IML models. Got: [{event.action_type}]'
            )

    def process_fs_event_src_code(self, event: FileSystemEvent):
        """
        Process FileSystem event when it's for a source code file
        """

        if event.action_type == FileSystemEventType.CREATED:
            if self.make_path_relative(event.abs_path1) in self.models:
                log.warning(
                    'Attempting to create a model with a path that already exists!'
                )
                return
            else:
                log.info(f'Creating a model for {event.abs_path1}')

                abs_path = event.abs_path1
                rel_path = self.make_path_relative(abs_path)
                try:
                    with open(abs_path) as infile:
                        srcCode = infile.read()
                except Exception:
                    log.error(f'Failed to read contents of the created file {abs_path}')
                    return

                self.models[rel_path] = Model(rel_path=rel_path, src_code=srcCode)
                log.info(f"We've now created a new model for path={rel_path}.")

        elif event.action_type == FileSystemEventType.MODIFIED:
            # A source file has been modified, so we will update our model info
            log.info(f'Processing modified source file: {event.abs_path1}')

            abs_path = event.abs_path1
            rel_path = self.make_path_relative(abs_path)

            try:
                with open(abs_path) as infile:
                    srcCode = infile.read()
            except Exception as e:
                log.error(
                    f'Failed to open a modified file: fullpath={abs_path}; error={str(e)}'
                )
                return

            if rel_path in self.models:
                # Let's now update the model's source code and record when we've done it
                self.models[rel_path].set_src_code(srcCode, True)
                log.info(
                    f"We've now updated source code for model with path=[{abs_path}]"
                )
            else:
                log.info(
                    'Modified file does not have a corresponding model. Creating it.'
                )
                self.models[rel_path] = Model(rel_path=rel_path, src_code=srcCode)
                log.info(f"We've now created a new model for path={abs_path}.")

        elif event.action_type == FileSystemEventType.DELETED:
            # We need to process a deleted file event
            log.info(f'Processing deleted source file: {event.abs_path1}')
            rel_path = self.make_path_relative(event.abs_path1)
            if rel_path in self.models:
                self.models[rel_path].set_src_code(None)
            else:
                log.error(
                    f'Attempting to process delete event for a non-existent model {event.abs_path1}'
                )

        elif event.action_type == FileSystemEventType.MOVED:
            # We need to process a deleted file event
            log.info(f'Processing a moved file: {event.abs_path1} to {event.abs_path2}')
            rel_path1 = self.make_path_relative(event.abs_path1)
            rel_path2 = self.make_path_relative(event.abs_path2)

            if rel_path1 in self.models and rel_path2 not in self.models:
                # self._models[rel_path].setSrcCode(None)
                self.models[rel_path2] = self.models[rel_path1]
                self.models[rel_path2].rel_path = rel_path2
                del self.models[rel_path1]
        else:
            log.error(
                "Something went wrong. We've gotten an unknown event type: {event.action_type}"
            )

    def run_file_sync(self):
        """
        Go through the source code directory and update the models.
        """

        log.info(f'Building FS from {self.src_dir_abs_path}')
        fs = FileSystem.from_disk(Path(self.src_dir_abs_path))

        py_dep = build_py_dep_graph(fs)

        # Let's look for deleted nodes
        model_paths = {str(m.relative_path) for m in py_dep.nx_graph}
        for path in self.models:
            log.info(f'Checking path {path}...')
            if path not in model_paths:
                # This should indicate that the model has since been deleted
                self.models[path].set_src_code(None)

        # Now let's look at new nodes update code for the existing ones
        for node in py_dep.nodes:
            rel_path = str(node.relative_path)
            if rel_path not in self.models:
                log.info(f'Adding model for path {rel_path}')
                model = Model(
                    rel_path=rel_path,
                    src_code=node.content,
                    src_language=self.language,
                )
                self.models[rel_path] = model
                log.info(f'Creating new model with path={node.relative_path}.')
            else:
                log.info(f'Setting source code to existing node={node.relative_path}.')
                self.models[rel_path].set_src_code(node.content)

        # Now let's add the dependencies
        for p1, p2 in py_dep.nx_graph.edges:
            path1, path2 = str(p1.relative_path), str(p2.relative_path)
            if path1 in self.models and path2 in self.models:
                model1 = self.models[path1]
                model2 = self.models[path2]
                if model2 not in model1.dependencies:
                    model1.dependencies.append(model2)

            else:
                errorMsg = "Attempting to add a dependency that doesn't exist!"
                log.error(errorMsg)
                raise Exception(errorMsg)

        # We should now have our dict with models representing all the files in the
        # directory right now and those that have been deleted!
        log.info('FileSync ran successfully!')

    def apply_sketch(self, sketch: Sketch) -> None:
        """
        Apply sketch back to the MetaModel
        """

        raise NotImplementedError('Sketch application is not implemented yet!')

    @staticmethod
    def fromJSON(j: str | dict):
        """
        Return an object from JSON file
        """

        if isinstance(j, str):
            return MetaModel.model_validate_json(j)
        elif isinstance(j, dict):
            return MetaModel.model_validate(j)
        else:
            raise Exception('Input must be either a str or a dict!')

    def toJSON(self):
        """
        Return a JSON object
        """
        return self.model_dump_json()

    def summary(self):
        """Return summary of the meta model state"""
        allFormStatuses = [
            FormalizationStatus.UNKNOWN,
            FormalizationStatus.TRANSPARENT,
            FormalizationStatus.INADMISSIBLE,
            FormalizationStatus.EXECUTABLE_WITH_APPROXIMATION,
            FormalizationStatus.ADMITTED_WITH_OPAQUENESS,
        ]

        frmStatuses = {s: 0 for s in allFormStatuses}
        for model in self.models.values():
            frmStatuses[model.formalization_status()] += 1

        def count(f):
            return sum(len(f(m)) for m in self.models.values())

        return {
            'frm_statuses': frmStatuses,
            **{
                label: count(f)
                for label, f in [
                    ('num_models', lambda _: [None]),
                    ('num_vgs', lambda m: m.verification_goals()),
                    ('num_opaque_funcs', lambda m: m.opaque_funcs()),
                    ('num_decomps', lambda m: m.decomps()),
                ]
            },
        }

    def rich_summary(self):
        """Return rich summary of the meta model state"""
        summary = self.summary()

        def frm_status(status_count):
            status, count = status_count
            return Text.assemble(status.__rich__(), ' ', (str(count), 'bold'))

        def fmt_countable(label_key):
            label, key = label_key
            return Text.assemble(f'Num. {label}: ', (str(summary[key]), 'bold'))

        countables = [
            ('Models', 'num_models'),
            ('Verification goals', 'num_vgs'),
            ('Opaque functions', 'num_opaque_funcs'),
            ('Decompositions', 'num_decomps'),
        ]
        return Group(
            Text.assemble(*intersperse(' | ', map(fmt_countable, countables))),
            Text.assemble(
                *intersperse(' | ', map(frm_status, summary['frm_statuses'].items()))
            ),
        )

    def __repr__(self):
        """Return a nice representation"""

        metaSummary = self.summary()
        statuses = {s.name: n for s, n in metaSummary['frm_statuses'].items()}

        s = f'Source file directory: {self.src_dir_abs_path}\n'
        s += f'Number of models: {metaSummary["num_models"]}\n'
        s += f'Number of VGs: {metaSummary["num_vgs"]}\n'
        s += f'Number of decomps: {metaSummary["num_decomps"]}\n'
        s += f'Status breakdown: \n{json.dumps(statuses, indent=4)}'
        s += '\n'
        s += 'Model details:\n'

        for path in self.models:
            s += f'Model [path={path}; {self.models[path].str_summary()}]\n'

        return s

    def __rich__(self):
        """Return RICH values"""
        pretty = Pretty(self.toJSON(), indent_size=2)
        return Panel(pretty, title='Current MetaModel', border_style='green')


class MetaModelUtils:
    """Some helper functions for reasoning about metamodels"""

    @staticmethod
    def modelContentsChanged(mmodel1: MetaModel, mmodel2: MetaModel):
        if set(mmodel1.models.keys()) != set(mmodel2.models.keys()):
            return True

        # TODO Add other conditions, like changes to opaque functions, etc...
        return False



#
#   Imandra Inc.
#
#   cmd_search.py
#
from logging import getLogger

from rich import print as rprint

from .utils import CLServerClient

log = getLogger(__name__)


def run_search(query: str, addr: str):
    """
    Run search command
    """
    cl_client = CLServerClient(addr)
    response = cl_client.get('search', {'query': query})

    if len(response.json()) == 0:
        print('No results found!')
    else:
        for res in response.json():
            rprint(res)



#
#   Imandra Inc.
#
#   cmd_models.py
#

import time
from logging import getLogger
from typing import Annotated

import typer
from rich import print as printr
from rich.live import Live

from ..strategy.model import Model, ModelList
from .utils import CLServerClient

log = getLogger(__name__)
app = typer.Typer()


# fmt: off
@app.command('list')
def run_model_list(
    live: Annotated[bool, typer.Option(help='If used, then will run the command in a loop.')] = False,
    addr: Annotated[str, typer.Option(help='CL Server address')] = 'http://127.0.0.1:8000',
):
    # fmt: on
    """
    List info about all models in the current strategy
    """

    cl_client = CLServerClient(addr)
    resp = cl_client.get('metamodel/list')
    model_list = ModelList.model_validate_json(resp.json())

    if live:
        with Live(model_list, refresh_per_second=4) as l:
            while True:
                resp = cl_client.get('metamodel/list')
                l.update(ModelList.model_validate_json(resp.json()))
                time.sleep(1)

    else:
        printr(model_list)


# fmt: off
@app.command('view')
def run_model_view(
    index: Annotated[int, typer.Argument(help='Index of the model to view')],
    live: Annotated[bool, typer.Option(help='If used, then will run the command in a loop.')] = False,
    addr: Annotated[str, typer.Option(help='CL Server address')] = 'http://127.0.0.1:8000',
):
    # fmt: on
    """
    View model specified by index
    """

    cl_client = CLServerClient(addr)
    resp = cl_client.get(f'model/byindex/{index}')
    model = Model.model_validate_json(resp.json())

    if live:
        with Live(model, refresh_per_second=1) as l:
            resp = cl_client.get(f'model/byindex/{index}')
            l.update(Model.model_validate_json(resp.json()))
            time.sleep(1)
    else:
        printr(model)


# fmt: off
@app.command('freeze')
def run_model_cmd_freeze(
    index: Annotated[int, typer.Argument(help='Freezes the IML code b/c of user changes.')],
    addr: Annotated[str, typer.Option(help='CL Server address')] = 'http://127.0.0.1:8000',
):
    # fmt: on
    cl_client = CLServerClient(addr)
    resp = cl_client.post(f'model/cmd/freeze/{index}')
    printr(resp.json())


@app.command('unfreeze')
def run_model_cmd_unfreeze(
    index: Annotated[
        int,
        typer.Argument(
            help='Unfreezes the model if frozen. User-specified IML code will be overriden if needed.'
        ),
    ],
    addr: Annotated[str, typer.Option(help='CL Server address')] = 'http://127.0.0.1:8000',
):
    cl_client = CLServerClient(addr)
    resp = cl_client.post(f'model/cmd/unfreeze/{index}')
    printr(resp.json())



#
#   Imandra Inc.
#
#   cmd_sketches.py
#

from logging import getLogger
from typing import Annotated

import typer
from rich import print as rprint

from .utils import CLServerClient

log = getLogger(__name__)
app = typer.Typer()


@app.command('list')
def run_sketches_list(
    addr: Annotated[
        str, typer.Option(help='CL Server address')
    ] = 'http://127.0.0.1:8000',
):
    """
    Run search command
    """

    cl_client = CLServerClient(addr)
    response = cl_client.get('sketches/list')

    if len(response.json()) == 0:
        print('No results found!')
    else:
        for res in response.json():
            rprint(res)


# fmt: off
@app.command('create')
def run_sketches_create(
    rel_path: Annotated[str, typer.Argument(help='Relative path of the model to use as the anchor')],
    addr: Annotated[str, typer.Option(help='CL Server address')] = 'http://127.0.0.1:8000',
):
    # fmt: on
    """
    Create a new sketch
    """
    typer.echo("I'm in run_sketches_create")


# fmt: off
@app.command('delete')
def run_sketches_delete(
    sketch_id: Annotated[str, typer.Argument(help='Sketch ID of the sketch to be deleted')],
    addr: Annotated[str, typer.Option(help='CL Server address')] = 'http://127.0.0.1:8000',
):
    # fmt: on
    """
    Delete a sketch
    """
    typer.echo("I'm in run_sketches_delete")



import asyncio

from imandra.u.agents import create_thread_sync, get_remote_graph
from imandra.u.agents.code_logician.command import (
    EmbedCommand,
    InitStateCommand,
)
from imandra.u.agents.code_logician.graph import GraphState
from strategy.model import Embedding

src_code = """
def f(a, b):
    return a + b
"""

iml_code = """
let f a b = a + b
"""

graph = get_remote_graph(
    'code_logician', api_key='29U5z4uV1E1Jbg6bOzdT4kpJUoqSKgaoaVzlGyt1zQfNXjFd'
)
create_thread_sync(graph)
gs = GraphState()
gs = gs.add_commands(
    [
        InitStateCommand(src_code=src_code, src_lang='python'),
        EmbedCommand(),
    ]
)

gs1, _ = asyncio.run(gs.run(graph))

res = gs1.steps[-1].message
assert res is not None

embedding_metadata = res['metadata']
src_embedding = res['src_embeddings']

for e in res['embedding']:
    ee = Embedding()

print(res['metadata'])
print(src_embedding)

# 3072
print()
print()
print(f'length is {len(src_embedding[0]["embedding"])}')
print()
print()



#
#   Imandra Inc.
#
#   dep_graph.py
#


from .model import Model
from .model_status import SrcCodeStatus


class DependencyGraph:
    """Set of utilities for checking that there are no cycles in the models"""

    @staticmethod
    def checkNoCycles(models: dict[str, Model]):
        """Returns True if there're no loops, False otherwise"""

        return True

    @staticmethod
    def setDependencyChanges(model: Model, models: dict[str, Model]):
        """For a given model, go through everything that it affects and change depsChanges to True for it"""

        for path, m in models.values():
            if path == model.rel_path():
                continue

            if model in m.dependencies():
                m.setDependenciesChanged(True)

                DependencyGraph.setDependencyChanges(m, models)

    @staticmethod
    def getAffected(model: Model, existing_list: list[Model] = []):
        """Return the list of paths for all models ultimately affected by changes to this model"""

        for m in model.affects():
            pass

        return existing_list

    @staticmethod
    def identifyNextBatch(models: dict[str, Model]):
        """Given the various models -> figure out which ones we can work on next"""

        pass


def testOne():
    """testOne"""

    m1 = Model('a.py', 'hello')
    m2 = Model('b.py', 'hello')
    m3 = Model('c.py', 'Hello')

    models = {'a.py': m1, 'b.py': m2, 'c.py': m3}

    m1.dependencies(m2)
    m1.dependencies(m3)

    m2.dependencies(m3)

    for _, m in models.values():
        print(m)

    print("Now let's change the model and update its depdendencies")

    # This should now make the model require formalization task to be performed
    m3.setFileSyncStatus(SrcCodeStatus.SRC_CODE_CHANGED)
    DependencyGraph.setDependencyChanges(m3, models)

    # The depdendencies should be updated now...
    for _, m in models.values():
        print(m)


if __name__ == '__main__':
    try:
        testOne()
    except Exception as e:
        print(f'Failed testOne: {str(e)}')



from __future__ import annotations

from enum import IntEnum
from pathlib import Path
from typing import ClassVar, Literal, Self

import networkx as nx
from pydantic import BaseModel, Field, model_validator

from ..tools.base import (
    DepGraph,
    NodeMixin,
    nx_graph_from_adjacency_list,
)

# from ..tools.filesystem import Dir, File, FileSystem


class SymbolKind(IntEnum):
    """A symbol kind.
    Imported from multilspy.lsp_protocol_handler.lsp_types.SymbolKind.
    """

    File = 1
    Module = 2
    Namespace = 3
    Package = 4
    Class = 5
    Method = 6
    Property = 7
    Field = 8
    Constructor = 9
    Enum = 10
    Interface = 11
    Function = 12
    Variable = 13
    Constant = 14
    String = 15
    Number = 16
    Boolean = 17
    Array = 18
    Object = 19
    Key = 20
    Null = 21
    EnumMember = 22
    Struct = 23
    Event = 24
    Operator = 25
    TypeParameter = 26


Range = tuple[tuple[int, int], tuple[int, int]]


class Symbol(BaseModel, NodeMixin):
    """A mix of LSP's SymbolInformation and DocumentSymbol."""

    name: str
    relative_path: Path = Field()
    kind: SymbolKind
    range: Range | None = Field(
        default=None,
        description='The range of the symbol. None represents a file-level symbol.',
    )
    detail: str | None = Field(default=None)
    container_name: str | None = Field(default=None, alias='containerName')
    children: list[Symbol] = Field(default_factory=list)

    @property
    def path_key(self) -> str:
        return str(self.relative_path)

    def __repr__(self) -> str:
        return str(self.relative_path)


class Module(Symbol):
    """A module."""

    name: str | None = Field(
        default=None,
        description='The name of the module. Automatically generated from the relative '
        'path for Python and IML modules.',
    )
    kind: Literal[SymbolKind.Module] = Field(default=SymbolKind.Module)
    content: str
    src_lang: str

    def __hash__(self) -> int:
        return hash((self.relative_path, self.content))

    @model_validator(mode='after')
    def validate_path(self) -> Self:
        if self.src_lang == 'python' and not self.relative_path.suffix == '.py':
            raise ValueError('Python module must have a .py extension')
        elif self.src_lang == 'iml' and not self.relative_path.suffix == '.iml':
            raise ValueError('IML module must have a .iml extension')
        return self


class ModuleDep[T: Module](DepGraph[T]):
    FILE_EXTENSION_MAP: ClassVar[dict[str, str]] = {
        'python': '.py',
        'iml': '.iml',
    }

    def __init__(self, nx_graph: nx.DiGraph):
        # Check if all nodes have the same source language
        src_langs = {node.src_lang for node in nx_graph.nodes}
        if len(src_langs) != 1:
            raise ValueError(
                f'All nodes must have the same source language, but got {src_langs}'
            )
        # Check if the source language is supported
        src_lang = next(iter(src_langs))
        if src_lang not in self.FILE_EXTENSION_MAP:
            raise ValueError(
                f'Unsupported source language: {src_lang}. Supported languages: '
                f'{list(self.FILE_EXTENSION_MAP.keys())}'
            )

        self.src_lang = src_lang
        super().__init__(nx_graph)

    @classmethod
    def from_adjacency_list(
        cls,
        nodes: list[T],
        adjacency_list: dict[str, set[str]],
    ) -> Self:
        nx_graph = nx_graph_from_adjacency_list(
            nodes=nodes, adjacency_list=adjacency_list
        )
        return cls(nx_graph)

    # FIXME: unused and does not type check (no body attribute in Module)
    # def into_file_system(
    #     self,
    #     repo_path: Path,
    #     filter_empty: bool = True,
    # ) -> FileSystem:
    #     nodes = self.nodes

    #     src_lang = self.src_lang
    #     file_extension = self.FILE_EXTENSION_MAP[src_lang]
    #     ignore_patterns = None

    #     files, dirs = [], []
    #     for node in nodes:
    #         rel_path = node.relative_path
    #         if filter_empty:
    #             if src_lang == 'iml' and not node.body.strip():
    #                 continue
    #             elif src_lang == 'python' and not node.content.strip():
    #                 continue  #

    #         files.append(
    #             File(
    #                 relative_path=rel_path,
    #                 content=node.content,
    #             )
    #         )
    #         for dir_path in rel_path.parents:
    #             dir = Dir(relative_path=dir_path)
    #             if dir not in dirs:
    #                 dirs.append(dir)

    #     return FileSystem.from_nodes(
    #         nodes=files + dirs,
    #         repo_path=repo_path,
    #         file_extensions=[file_extension],
    #         ignore_patterns=ignore_patterns,
    #     )



#
# Imandra Inc.
#
# endpts_search.py
#

import logging

from fastapi import HTTPException

from ..tools.cl_caller import calc_search_embeddings
from .cl_server import CLServer
from .search import SearchResult

log = logging.getLogger(__name__)


def register_search_endpoints(app: CLServer):
    """
    Attach search functions to the server.
    """

    @app.get('/search', operation_id='get_server_status')
    async def search(query: str) -> list[SearchResult]:
        """
        Returns the results of searching the across all the strategies.
        """
        try:
            query_emb = await calc_search_embeddings(query)
        except Exception as e:
            errMsg = f'Failed to calculate query embeddings: {e}'
            log.error(errMsg)
            raise HTTPException(status_code=404, detail=errMsg)

        if query_emb is None:
            log.warning('calc_search_embeddings was returned as None')
            return []

        # let's first get the embeddings vector and then we'll do something special...

        res = app.search(query_emb)
        log.info(f'Search result is {res}')

        return res



#
#   Imandra Inc.
#
#   cmd_strategy.py
#
from logging import getLogger
from typing import Annotated

import typer
from rich import print as printr
from rich.console import Console
from rich.pretty import Pretty

from .utils import CLServerClient

log = getLogger(__name__)
app = typer.Typer()

console = Console()


@app.command(name='list')
def run_strategy_list():
    """
    List all current strategies
    """
    cl_client = CLServerClient()
    resp = cl_client.get('/strategy/list').json()

    if len(resp) == 0:
        console.print('[bold magenta]No strategies[/bold magenta]')
    else:
        for s in resp:
            console.print_json(data=s)


@app.command(name='cws', help='Get current working strategy')
def run_strategy_cws():
    """
    Get the current working strategy
    """
    cl_client = CLServerClient()
    resp = cl_client.get('/strategy/cws')
    printr(f'Current working strategy: {resp}')


@app.command(name='setcws', help='Set current working stratey (cws)')
def run_strategy_set_cws(strat_id: Annotated[str, typer.Argument()]):
    """
    Set current working strategy
    """
    cl_client = CLServerClient()

    try:
        res = cl_client.post(f'/strategy/setcws/{strat_id}').json()
    except Exception as e:
        printr(f'Encountered an error during POST request: {e}')

    if 'detail' in res:
        printr(f'[bold red]{res["detail"]}[/bold red]')
    else:
        printr(
            f'[bold][magenta]Successfully updated current working strategy to: {strat_id}[/magenta][/bold]'
        )


@app.command(name='create', help='Create a new strategy within the server')
def run_strategy_create(
    stype: Annotated[str, typer.Argument(help="Strategy type (e.g. 'PyIML')")],
    spath: Annotated[str, typer.Argument(help='Strategy directory path')],
):
    """
    Create a new strategy
    """
    cl_client = CLServerClient()

    try:
        resp = cl_client.post(
            '/strategy/create', data={'strat_type': stype, 'strat_path': spath}
        )
    except Exception as e:
        printr(f'Encountered an error during POST request: {e}')
        return

    printr(f'[bold magenta]Successfully created a strategy: {resp}')


@app.command(name='delete', help='Delete/remove strategy from the server')
def run_strategy_delete(
    strat_id: Annotated[
        str, typer.Argument(help='Strategy ID of the strategy to be deleted')
    ],
):
    """
    Delete an existing strategy
    """
    cl_client = CLServerClient()

    try:
        resp = cl_client.post(f'/strategy/delete/{strat_id}')
    except Exception as e:
        printr(f'Encountered an error during POST request: {e}')
        return

    if resp == {'status': 'OK'}:
        printr(
            f'[bold magenta]Successfully deleted: {strat_id} strategy[/bold magenta]'
        )
    else:
        print(resp)
        printr('[bold red]Something went wrong[/bold red]')


@app.command(name='summary', help="Get current working strategy's summary")
def run_strategy_summary():
    """
    Retrieve current working strategy's summary
    """
    cl_client = CLServerClient()
    resp = cl_client.get('/strategy/summary').json()

    print('Summary for the current strategy:')
    printr(Pretty(resp))



#
#   Imandra Inc.
#
#   javaiml_strategy.py
#

import logging

from ..server.events import FileSystemEvent
from .pyiml_strategy import PyIMLStrategy
from .state import StrategyState

log = logging.getLogger(__name__)


class JavaIMLStrategy(PyIMLStrategy):
    """Java strategy"""

    _language = 'Java'

    def __init__(self, state: StrategyState):
        super().__init__(state=state)

    def watch_directories(self):
        """Return the list of directories the observer should watch"""
        return []
        # return self.abs_src_dir

    def on_load(self):
        """What should be done on startup"""
        pass

    def on_filesystem_event(self, event: FileSystemEvent):
        """ """
        log.info(f'Received {event} to process')

    def on_save(self):
        """
        What should be done on strategy save
        """
        pass



#
#   Imandra Inc.
#
#   model.py
#

import datetime
import logging
import re
from collections import deque
from pathlib import Path
from typing import Any

import numpy as np
from imandra.u.agents.code_logician.base import (
    VG,
    FormalizationDependency,
    FormalizationStatus,
    ModuleInfo,
    RegionDecomp,
    TopLevelDefinition,
)
from pydantic import BaseModel, field_serializer
from rich import box
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text

from codelogician.util import maybe_min

from ..util import filter, map
from .cl_agent_state import CLAgentState, Embedding, frm_status_to_rich
from .model_status import ModelStatus, SrcCodeStatus
from .model_task import FormalizationNeed, ModelTask, UserManualIMLEditTask

log = logging.getLogger(__name__)


class UserIMLEdit(BaseModel):
    """
    We'll just use this to keep track of user edits
    """

    user_iml_entry: str
    src_code_at_the_time: str
    time_of_edit: datetime.datetime


class Model(BaseModel):
    """
    Container for a model - representing a single model along with its dependendencies,
    formalization state, user-overriden IML code, etc...

    These are the primary events that the model handles:
    - 1. Update to the source code (will trigger CL agent state)
    - 2. Update to the IML code (by the user)
    - 3. Applying the result of running CL agent
    - 4. User freezes/unfreezes their IML code (to prevent it from
        being overriden by CL results)
    - 5. User resets the changes that he made to the model (to the last CL model available)

    The metamodel updates the model when those events occur (listed above). It then
    uses the Model status function to figure out wether a new event should be written
    or an update to IML code on disk

    Three ways that a status is accessed:
    - formalization_reasons(user_wait_time : int) -- returns a list of formalization reasons
      (why a CL task should be generated). The list may be comprised of the following:
        - FormalizationNeed.NO_AGENT_STATE - no formalization state exists
        - FormalizationNeed.SRC_CODE_CHANGED - underlying source code changed
        - FormalizationNeed.CONTEXT_ADDED - human has provided context for the model
        - FormalizationNeed.DEPS_CHANGED - formalization state of one or more of the dependencies changed

        -- if the returned list is empty, then no further action is required

    - iml_on_fs_update_ready() - should we update the iml model that's written to disk
        - Description: this can be set when we have an update to the model

    """

    rel_path: str
    src_code: str | None = None
    src_code_last_changed: datetime.datetime | None = None
    src_language: str = 'Python'

    # This contains the formalization state -
    # note it is updated by both - running CL on the source code changes (e.g. python)
    # and on user-manual edits to IML (we call a separate step for that)
    agent_state: CLAgentState | None = None

    # Here we account for user's changes to the model IML code
    # and keep track of how long it's been since they edited it
    # if it's been long enough and it's different that what we have
    # then we'll kick off a command to make it part of the Graph State
    user_iml_edit: UserIMLEdit | None = (
        None  # This is human provided IML code and timestamp
    )

    # If the user IML code is frozen, then even if the
    # CL comes up with an update to the model, then it would still not
    # override the file on disk
    iml_code_frozen: bool = False

    # Here we keep a record of what should be on disk
    # Because we have two sources of IML models on disk (agent and user)
    # we keep this here to check for differences if there's an update (i.e.
    # if we write the model to disk after CL results come in, then we update this
    # and when the FS event comes in for IML change, we disregard it)
    expected_iml_on_disk: str | None = None

    src_code_embeddings: list[Embedding] = []  # Embeddings for the original source code
    iml_code_embeddings: list[Embedding] = []  # IML code embeddings

    # This is human-provided context that is sent to the CL agent
    context: str | None = None

    # This is used to keep track of the latest TaskID that was generated
    # We can only apply changes to the CL agent state if the taskID of the
    # result andf what we're waiting for matches - otherwise we discard the results
    outstanding_task_ID: str | None = None

    # This will be typically inserted after creation of the object because
    # this requires actual object references (to other models)
    dependencies: list['Model'] = []  # list of models that this model depends on
    rev_dependencies: list['Model'] = []  # list of models that this model affects

    @field_serializer('dependencies', 'rev_dependencies')
    def serialize_tags(self, deps: list['Model']) -> list[str]:
        return map(lambda x: x.rel_path, deps)

    # This gets updated along with each agent state - here we maintain
    # IML code for the dependencies that were used. If they differ
    # then we may need to re-run CodeLogician.
    formalized_deps: list[FormalizationDependency] = []

    # We do the same with the human-provided context
    formalized_context: str | None = None

    # Static formalization reasons - note if the model is "live"
    # (i.e. not post deserialization (so we have actual dependencies present)
    # then this is not used - this is only to store those for serialization
    # when we have to present the model
    static_frm_reasons: list[FormalizationNeed] = []

    @field_serializer('static_frm_reasons')
    def serialize_frm_reasons(self, _):
        return self.formalization_reasons()

    def add_dependency(self, d: 'Model') -> None:
        """
        "Safe" add of a dependency - if it's already there, it would not add it twice
        """
        if d in self.dependencies:
            return
        self.dependencies.append(d)

    def status(self) -> ModelStatus:
        """
        Return a status object that contains various bits of the current state of the model
        """

        return ModelStatus(
            src_code_status=self.src_code_status(),
            instructions_added=self.context_provided(),
            deps_changed=self.deps_changed(),
            deps_need_formalization=self.deps_need_formalization(),
            outstanding_task_ID=self.outstanding_task_ID,
            formalization_status=self.formalization_status(),
        )

    @staticmethod
    def calc_distance(vec1, vec2) -> float:
        """
        Calculate cosine distance between two vectors
        """
        dot_product = np.dot(vec1, vec2)

        norm_vec1 = np.linalg.norm(vec1)
        norm_vec2 = np.linalg.norm(vec2)

        # Handle cases where one or both vectors are zero vectors
        # fmt: off
        return (1.0 if norm_vec1 == 0 or norm_vec2 == 0 else
                float(1 - dot_product / (norm_vec1 * norm_vec2)))
        # fmt: on

    def calc_embedding_distance(
        self, query_vector: list[float]
    ) -> None | tuple[float, str]:
        """
        Calculate distance between the specified query vector and its own embeddings.
        Return the one with smallest distance.
        """
        qv = np.array(query_vector)

        def min_dist(embeddings) -> None | float:
            distances = [self.calc_distance(qv, np.array(e.vector)) for e in embeddings]
            return maybe_min(distances)

        min_dist_iml = min_dist(self.iml_code_embeddings)
        min_dist_src = min_dist(self.src_code_embeddings)

        # fmt: off
        match min_dist_iml, min_dist_src:
            case None, None:         return None
            case None, min_dist_src: return min_dist_src, 'SRC'
            case min_dist_iml, None: return min_dist_iml, 'IML'
            case min_dist_iml, min_dist_src:
                return ((min_dist_iml, 'IML') if min_dist_iml < min_dist_src
                        else (min_dist_src, 'SRC'))
        # fmt: on

    def needs_embedding_update(self) -> bool:
        """
        Does this model need to update embedding info?
        """
        # TODO check to see if the source code/etc changed since last embedding...
        return False

    def gen_iml_user_update_task(self) -> ModelTask | None:
        """
        Generate a model task for updating the IML code within the graph state.
        What happens is that during the update, extract the VGs/Decomps, etc and place them
        in the right format within the graph state.
        """
        if self.user_iml_edit is None:
            return None

        def remove_import_directives(text: str) -> str:
            """
            Removes lines of the form [@@@import ...] from the given text.
            """
            return re.sub(r'^\s*\[@@@import.*?\]\s*\n?', '', text, flags=re.MULTILINE)

        task = UserManualIMLEditTask(
            rel_path=self.rel_path,
            graph_state=self.agent_state.graph_state if self.agent_state else None,
            iml_code=remove_import_directives(self.user_iml_edit.user_iml_entry),
            dependencies=self.dependencies_iml_models(),
        )

        self.outstanding_task_ID = task.task_id

        return task

    def gen_embedding_task(self) -> ModelTask:
        """
        Create an embedding task - TODO: we should streamline this
        """
        task = ModelTask(
            rel_path=self.rel_path,
            src_code=self.src_code if self.src_code else '',
            context=self.context,
            gen_embeddings=True,
        )

        # Let's now save this task so we know that we should expect
        # the result
        self.outstanding_task_ID = task.task_id

        return task

    def gen_formalization_task(self) -> ModelTask:
        """
        Generate a formalization task (ModelTask) object if we need to
        """

        task = ModelTask(
            rel_path=self.rel_path,
            src_code=self.src_code if self.src_code else '',
            context=self.context,
            graph_state=self.agent_state.graph_state if self.agent_state else None,
            dependencies=self.dependencies_iml_models(),
        )

        self.outstanding_task_ID = task.task_id

        return task

    def set_iml_code_frozen(self) -> None:
        """
        Set IML code frozen - this will prevent any
        """
        self.iml_code_frozen = True

    def set_iml_code_unfrozen(self) -> None:
        """
        Here we set the IML code unfrozen
        """
        self.iml_code_frozen = False

    def cl_result_available(self) -> bool:
        """
        Is there a CL result that has been available, but not set
        b/c of the frozen client model?
        """
        return False

    def context_provided(self) -> bool:
        """
        Return True if there's human-provided context
        """
        return bool(self.context)

    def iml_on_fs_update_ready(self) -> bool:
        """
        Returns True/False if there's an update to the IML model that should be written to disk...
        """
        return False

    def apply_agent_state(
        self,
        agent_state: CLAgentState,
        dependencies: list[FormalizationDependency] = [],
        result_of_user_edit: bool = False,
    ) -> None:
        """
        Apply the agent state and store the dependencies that were used in the call to CL agent

        Parameters:
        - agent_state - CLAgentState that we'll apply
        - dependencies - the dependencies that were used during formalization - we'll use them later to compare if
            we need to redo autoformalization.
        - result_of_user_edit : bool

        """

        # Let's make sure we have an actual CLAgentState
        if not isinstance(agent_state, CLAgentState):
            raise Exception(
                f'Expected a value of CLAgentState, but got something else: {str(agent_state)}!'
            )

        # Let's also now reset the task ID
        self.outstanding_task_ID = None

        # Agent state
        self.agent_state = agent_state

        # Remove instructions that were used from the current list of outstanding ones
        self.formalized_context = agent_state.context

        # Let's now set the dependencies' models that were used
        self.formalized_deps = dependencies

        # We can also update the embeddings
        self.src_code_embeddings = agent_state.src_code_embeddings
        self.iml_code_embeddings = agent_state.iml_code_embeddings

    def user_iml_change_ready(self, user_wait_time: int | None) -> bool:
        """
        Is the user's IML change ready to be turned into a task. Here, we consider whether
        the change done by the user is "old enough". We do this to ensure we're only formalizing
        changes that have become 'stable'.
        """

        if self.user_iml_edit is None:
            return False

        if user_wait_time is None:
            long_enough = True
        else:
            long_enough = (
                datetime.datetime.now() - self.user_iml_edit.time_of_edit
            ) > datetime.timedelta(seconds=user_wait_time)

        return long_enough

    def formalization_reasons(
        self, user_wait_time: int | None = None
    ) -> list[FormalizationNeed]:
        """
        Unlike the function `user_iml_change_ready`, here we only look at the source code changes.
        Returns a list of reasons for the need to perform formalization on this model.
        If the list is empty - then there's no need to do it.
        """

        if self.src_code is None:
            return []  # if we have no source code (i.e. file was deleted), then there's nothing to do
        if self.agent_state is None:
            return [FormalizationNeed.NO_AGENT_STATE]

        reasons: list[FormalizationNeed] = []
        if self.has_src_code_changed(user_wait_time):
            reasons.append(FormalizationNeed.SRC_CODE_CHANGED)
        if self.context_provided():
            reasons.append(FormalizationNeed.CONTEXT_ADDED)
        if self.deps_changed():
            reasons.append(FormalizationNeed.DEPS_CHANGED)

        return reasons

    def deps_need_formalization(self, user_wait_time: int = 0):
        """
        Return True if there're dependencies (may be indirect) that are due for formalization and
        False otherwise. We use this function when constructing the list of Tasks for the metamodel.
        """

        for d in self.dependencies:
            if d.formalization_reasons(user_wait_time) or d.deps_need_formalization(
                user_wait_time
            ):
                return True

        return False

    def full_deep_dependencies(self) -> list['Model']:
        """
        Get a full deep list of dependencies
        """

        result = []

        def helper(m: 'Model'):
            """
            Traverse the dependencies graph
            """

            for m in m.dependencies:
                if m not in result:
                    result.append(m)

                helper(m)

        helper(self)

        return result

    def shallow_dep_iml_models(self) -> list[FormalizationDependency]:
        """
        Return the list of only the top-level
        """

        if len(self.dependencies) == 0:
            return []

        ds: list[FormalizationDependency] = []

        for model in self.dependencies:
            src_mod_info = model.to_CL_src_module_info()
            iml_mod_info = model.to_CL_iml_module_info()

            # We only care about the models that have IML code available
            if iml_mod_info is None:
                continue

            ds.append(
                FormalizationDependency(
                    src_module=src_mod_info,
                    iml_module=iml_mod_info,
                )
            )

        return ds

    def dependencies_iml_models(self) -> list[FormalizationDependency]:
        """
        This is used in two places -> to print out the full model and to
        generate formalizaiton task context
        """

        log.debug(f'Will now get the list of dependencies for {self.rel_path}')

        if len(self.dependencies) == 0:
            return []

        ds: list[FormalizationDependency] = []

        full_list = self.full_deep_dependencies()

        sorted_list = Model.do_topological_sort(full_list)

        for model in sorted_list:
            src_mod_info = model.to_CL_src_module_info()
            iml_mod_info = model.to_CL_iml_module_info()

            # We only care about the models that have IML code available
            if iml_mod_info is None:
                continue

            ds.append(
                FormalizationDependency(
                    src_module=src_mod_info,
                    iml_module=iml_mod_info,
                )
            )

        log.debug('Done')

        return ds

    def make_dep_name_from_path(self, rel_path: str) -> str:
        """
        Create a name for a dependency
        """
        p = Path(rel_path).parent / Path(rel_path).stem

        return ('_'.join(Path(p).parts)).capitalize()

    def to_CL_src_module_info(self) -> ModuleInfo:
        """
        Create ModuleInfo
        """
        return ModuleInfo(
            name=self.make_dep_name_from_path(self.rel_path),
            relative_path=Path(self.rel_path),
            content=self.src_code if self.src_code else '',
            src_lang=self.src_language,
        )

    def to_CL_iml_module_info(self) -> ModuleInfo | None:
        """
        Return ModuleInfo object for the IML code if it's available, None otherwise
        """

        # If we don't have any IML code for this model, let's not return it
        if self.iml_code() is None:
            return None

        if self.iml_code() is not None:
            iml_code = str(self.iml_code())
        else:
            iml_code = ''

        return ModuleInfo(
            name=self.make_dep_name_from_path(self.rel_path),
            relative_path=Path(self.rel_path),
            content=iml_code,
            src_lang='IML',
        )

    def src_code_status(self, srcCodeWaitTime: int | None = None) -> SrcCodeStatus:
        """
        Return source code status
        """
        if self.src_code is None:
            return SrcCodeStatus.SRC_CODE_DELETED

        else:
            if srcCodeWaitTime is None or self.src_code_last_changed is None:
                timeLongEnough = True
            else:
                timeLongEnough = (
                    datetime.datetime.now() - self.src_code_last_changed
                ) > datetime.timedelta(seconds=srcCodeWaitTime)

        if timeLongEnough and (hash(self.src_code) == hash(self.iml_code())):
            return SrcCodeStatus.SRC_CODE_CHANGED
        else:
            return SrcCodeStatus.SRC_CODE_CURRENT

    def formalization_status(self) -> FormalizationStatus:
        """
        Return formalization status of the model
        """
        if self.agent_state is None:
            return FormalizationStatus.UNKNOWN
        else:
            return self.agent_state.status

    def has_src_code_changed(self, user_wait_time: int | None = None) -> bool:
        """
        Return True if source code has changed from when it was formalized
        """

        # if we're missing formalization altogether, then source code has changed
        if self.agent_state is None:
            return True

        if user_wait_time is None or self.src_code_last_changed is None:
            timeLongEnough = True
        else:
            timeLongEnough = (
                datetime.datetime.now() - self.src_code_last_changed
            ) > datetime.timedelta(seconds=user_wait_time)

        return timeLongEnough and (
            hash(self.src_code) != hash(self.agent_state.src_code)
        )

    def deps_changed(self) -> bool:
        """
        Return True if dependencies' models (IML code) have changed since we
        ran CL formalization task last time. When we receive CL results, we always
        record dependencies' IML models that were used.
        """
        currentDeps = self.shallow_dep_iml_models()
        paths_list = map(lambda x: Path(x.rel_path), self.dependencies)
        in_paths_list = lambda x: x.src_module.relative_path in paths_list  # noqa
        relevant_formalized_deps = filter(in_paths_list, self.formalized_deps)

        # This will iterate through the list and pick up any changes
        # to the dependencies (they're Pydantic model objects)
        return len(relevant_formalized_deps) != len(currentDeps) or any(
            d not in currentDeps for d in relevant_formalized_deps
        )

    def create_import_iml_statements(self):
        """
        Generate a list of import statements to the relevant IML models.
        """
        text = ''
        for d in self.dependencies_iml_models():
            name = d.iml_module.name
            rel_path = str(Path(d.iml_module.relative_path).with_suffix('.iml'))
            text += f'[@@@import {name}, "{rel_path}"]\n'

        return text

    def set_iml_model(self, new_iml_model: str, record_time: bool = False) -> None:
        """
        Set the IML model code and record the time if we need to. This is called
        when the user manually overrides the model. We also need to check if the model
        actually changed... (this event is all also  called when we write out the model
        from CL result - it can't tell the difference, so we need to check it here...)
        """

        # We use hash as it also handles the case when models are None
        if hash(new_iml_model.strip()) == hash(str(self.expected_iml_on_disk).strip()):
            log.warning(
                f"Model set by the user doesn't appear to be any different from existing model: [{self.rel_path}]"
            )
            return

        log.info(f'User is updating the IML model: [{self.rel_path}]')

        self.user_iml_edit = UserIMLEdit(
            user_iml_entry=new_iml_model,
            src_code_at_the_time='N\\' if self.src_code is None else self.src_code,
            time_of_edit=datetime.datetime.now(),
        )

    def set_src_code(self, new_src_code: str | None, record_time: bool = False) -> None:
        """
        Set the src_code and update the time source code changed
        """

        # This means the code was deleted
        if new_src_code is None:
            self.src_code = None
            self.src_code_last_changed = None
            return

        # if not isinstance(new_src_code, str):
        #    raise Exception (f"Expected a string value for `new_src_code`, but got {type(new_src_code).__name__}")

        if hash(new_src_code) != hash(self.src_code):
            log.info(f'Updating source code for model [{self.rel_path}]')

            if record_time:
                self.src_code_last_changed = datetime.datetime.now()
            else:
                self.src_code_last_changed = None

            self.src_code = new_src_code
        else:
            log.warning(f"Specified model hasn't changed: [{self.rel_path}]")

    def iml_code_to_print(self) -> str:
        """
        Provide IML code (if available) along with relevant import statements
        """
        if self.iml_code() is None:
            return 'N\\A'

        import_statements = self.create_import_iml_statements()
        return f'{import_statements}\n\n{self.iml_code()}'

    def iml_code(self) -> str | None:
        """
        If available, returns IML code with artifacts (e.g. VGs and decomps)
        """
        return self.agent_state.iml_code if self.agent_state else None

    def iml_model(self) -> str | None:
        """
        If available, returns IML model (not artifacts like VGs, decomps). This
        also looks at the available user-overriden code.
        """

        return self.agent_state.iml_model if self.agent_state else None

    def verification_goals(self) -> list[VG]:
        """
        If available, returns the list of dictionaries with verification goals
        """
        return self.agent_state.vgs if self.agent_state else []

    def failed_vgs(self) -> list[VG]:
        """
        Return just the list of failed VGs
        """
        return filter(lambda x: not x['status'], self.verification_goals())

    def decomps(self) -> list[RegionDecomp]:
        """
        If available, returns the list of decomposition requests
        """
        return self.agent_state.region_decomps if self.agent_state else []

    def opaque_funcs(self) -> list[TopLevelDefinition]:
        """
        If available, return the opaque functions used along with their approximations
        """
        return self.agent_state.opaque_funcs if self.agent_state else []

    def gen_stats(self) -> dict[str, str | int]:
        """
        Generate some numberical stats for this model
        """
        s: dict[str, str | int] = {}
        s['frm_status'] = str(self.formalization_status())
        s['num_opaques'] = len(self.opaque_funcs())
        s['num_failed_vgs'] = len(self.failed_vgs())

        return s

    def toJSON(self) -> str:
        """
        Return a dictionary we can save to disk
        """

        # This will save the reasons which we can only calculate if
        # we have "live" dependency models (not strings as is the case
        # post-deserialization)
        self.static_frm_reasons = self.formalization_reasons(5)

        log.debug(f'Adding static frm reasons: [{self.static_frm_reasons}]')

        return self.model_dump_json()

    @staticmethod
    def fromJSON(j: str | dict[str, str]) -> 'Model':
        """
        Return a Model object from provided JSON. Note that we do not set the lists
        of models for 'dependsOn' here because we don't have access to the
        actual model objects. This is done outside this state methods.
        """

        if isinstance(j, str):
            return Model.model_validate_json(j)
        else:
            return Model.model_validate(j)

    def str_summary(self) -> str:
        """
        Return a one-line string with high-level details
        """
        return f'[status={self.status()}]'

    def __hash__(self):
        """
        We need this so we can compare models
        """
        return hash(str(self.toJSON()))

    def __repr__(self):
        """
        Return a nice representation
        """
        dependencyPaths = map(lambda x: x.rel_path, self.dependencies)
        agentStateStr = 'None' if self.agent_state is None else self.agent_state.status

        s = f'Model: {self.rel_path} \n'
        s += f'{str(self.status())}\n'
        s += f'Formalization state: {agentStateStr}\n'
        s += f'Depends on: {dependencyPaths}\n'
        s += f'Source language: {self.src_language}\n'

        if self.src_code:
            s += f'Source code (condensed): \n {self.src_code[:100]}\n'
        else:
            s += 'Source code (condensed): \n None \n'

        s += f'Opaque funcs: {len(self.opaque_funcs())}\n'
        s += f'Decomps: {len(self.decomps())}\n'
        return s

    def __rich__(self):
        """
        Return a Rich representation
        """
        table = Table(
            title=Text(f'Model: {self.rel_path}', style='bold italic magenta'),
            box=box.MINIMAL_DOUBLE_HEAD,
            show_lines=True,
        )

        table.add_column('Name')
        table.add_column('Attribute')

        src_code_str = 'N/A' if self.src_code is None else self.src_code
        user_iml_code_str = (
            'N/A' if self.user_iml_edit is None else self.user_iml_edit.user_iml_entry
        )
        user_iml_code_last_change = (
            'N/A'
            if self.user_iml_edit is None
            else str(self.user_iml_edit.time_of_edit)
        )

        needs_frm_str = 'Yes' if self.static_frm_reasons == [] else 'No'
        frozen_str = 'Yes' if self.iml_code_frozen else 'No'

        iml_code_str = 'N/A' if not self.iml_code() else str(self.iml_code())

        eval_result_str = 'N/A' if not self.agent_state else self.agent_state.eval_res

        # fmt: off
        table.add_row('Relative path',              Text(self.rel_path, style='bold')                      )
        table.add_row('Formalization status',       frm_status_to_rich(self.formalization_status())        )
        table.add_row('Eval result',                Text(eval_result_str, style='bold')                    )
        table.add_row('Needs formalization?',       Text(needs_frm_str, style='bold')                      )
        table.add_row('Formalization reasons',      Text(str(self.static_frm_reasons))                     )
        table.add_row('Frozen?',                    Text(frozen_str)                                       )
        table.add_row('Src last changed',           Text(f'{self.src_code_last_changed}', style='bold')    )
        table.add_row('Src code',                   Syntax(src_code_str, 'python')                         )
        table.add_row('IML code',                   Syntax(iml_code_str, 'ocaml')                          )
        table.add_row('User IML code',              Syntax(user_iml_code_str, 'ocaml')                     )
        table.add_row('User IML code last changed', Text(f'{user_iml_code_last_change}', style='bold')     )
        table.add_row('Src Code Embeddings',        Text(f'{bool(self.src_code_embeddings)}', style='bold'))
        table.add_row('IML Code Ebeddings',         Text(f'{bool(self.iml_code_embeddings)}', style='bold'))
        table.add_row('Context',                    Text(f'{self.context}', style='bold')                  )
        table.add_row('Task ID',                    Text(f'{self.outstanding_task_ID}', style='bold')      )
        table.add_row('Deps',                       Text(f'{self.dependencies}', style='bold')             )
        table.add_row('Opaques',                    Text(f'{self.opaque_funcs()}', style='bold')           )
        table.add_row('Verification Goals',         Text(f'{self.verification_goals()}', style='bold')     )
        table.add_row('Decompositions',             Text(f'{self.decomps()}', style='bold')                )
        # fmt: on

        return table

    @staticmethod
    def do_topological_sort(models: list['Model']) -> list['Model']:
        """
        Return list of paths sorted topologically by dependencies
        """

        path_to_idx, idx_to_path = {}, {}
        edges = []
        idx = 0

        def add_to_lookup(idx, path):
            path_to_idx[path] = idx
            idx_to_path[idx] = path

        # Get all the edges
        for model in models:
            if model.rel_path not in path_to_idx:
                add_to_lookup(idx, model.rel_path)
                idx += 1

            for d in model.dependencies:
                if d.rel_path not in path_to_idx:
                    add_to_lookup(idx, d.rel_path)
                    idx += 1

                edges.append((path_to_idx[model.rel_path], path_to_idx[d.rel_path]))

        def constructadj(V, edges):
            adj = [[] for _ in range(V)]
            for u, v in edges:
                adj[u].append(v)
            return adj

        def topologicalSort(V, edges):
            adj = constructadj(V, edges)

            indegree = [0] * V

            # Calculate indegree of each vertex
            for u in range(V):
                for v in adj[u]:
                    indegree[v] += 1

            # Queue to store vertices with indegree 0
            q = deque([i for i in range(V) if indegree[i] == 0])

            result = []
            while q:
                node = q.popleft()
                result.append(node)

                for neighbor in adj[node]:
                    indegree[neighbor] -= 1
                    if indegree[neighbor] == 0:
                        q.append(neighbor)

            # Check for cycle
            if len(result) != V:
                raise Exception('Graph contains cycle!')

            return result

        model_by_path = {m.rel_path: m for m in models}
        sorted_idxs = topologicalSort(len(models), edges)
        return [model_by_path[idx_to_path[i]] for i in reversed(sorted_idxs)]


class ModelList(BaseModel):
    """
    We use this with terminal CLI tools
    """

    models: list[Model]

    # FIXME: this doesn't look right
    # def toJSON(self) -> list[str]:
    #     """
    #     Return a JSON
    #     """
    #     return [m.toJSON() for m in self.models]

    @staticmethod
    def fromJSON(j: str | dict) -> 'ModelList':
        if isinstance(j, str):
            return ModelList.model_validate_json(j)
        elif isinstance(j, dict):
            return ModelList.model_validate(j)
        else:
            raise Exception('Input must be either a str or a dict!')

    @staticmethod
    def get_model_summary(m: Model) -> dict[str, Any]:
        """
        Create a dict with various summary stats of the provided model
        """

        # TODO: this dictionary is only ever used in one place, in the the
        # __rich__ method below, so we can merge these two methods and avoid
        # building this weakly typed dictionary.
        return {
            'frm_status': m.formalization_status(),
            'needs_frm': 'No' if m.static_frm_reasons == [] else 'Yes',
            'frm_reasons': map(str, m.static_frm_reasons),
            'is_frozen': 'Yes' if m.iml_code_frozen else 'No',
            'rel_path': m.rel_path,
            'src_last_ched': m.src_code_last_changed,
            'user_iml_code': m.user_iml_edit is not None,
            'user_iml_last_ched': (
                str(m.user_iml_edit.time_of_edit) if m.user_iml_edit else 'N\\A'
            ),
            'src_code_embds': bool(m.src_code_embeddings),
            'iml_code_embds': bool(m.iml_code_embeddings),
            'context': m.context,
            'task_id': m.outstanding_task_ID,
            'deps': m.dependencies,
            'num_opaques': len(m.opaque_funcs()),
            'num_vgs': len(m.verification_goals()),
            'num_decomps': len(m.decomps()),
        }

    def __rich__(self):
        """
        We'll used this in the CLI
        """

        table = Table(
            title=Text('Models list', style='bold italic magenta'),
            box=box.MINIMAL_DOUBLE_HEAD,
            show_lines=True,
        )

        # fmt: off
        table.add_column('ID'                       , justify='right', no_wrap=True )
        table.add_column('Frm Status'               , justify='right', no_wrap=True )
        table.add_column('Frozen'                   , justify='right', no_wrap=True )
        table.add_column('Needs frm?'               , justify='right', no_wrap=True )
        table.add_column('Frm Needs'                , justify='right', no_wrap=False)
        table.add_column('Rel path'                 , justify='right', no_wrap=True )
        table.add_column('Src Last Changed'         , justify='right', no_wrap=True )
        table.add_column('User IML Code'            , justify='right', no_wrap=True )
        table.add_column('User IML Code Last Chged' , justify='right', no_wrap=True )
        table.add_column('Src Code Embds'           , justify='right', no_wrap=True )
        table.add_column('IML Code Embds'           , justify='right', no_wrap=True )
        table.add_column('Context'                  , justify='right' )
        table.add_column('Task ID'                  , justify='right' )
        table.add_column('Dependencies'             , justify='right', no_wrap=False)
        table.add_column('Num opaques'              , justify='right' )
        table.add_column('Num VGs'                  , justify='right' )
        table.add_column('Num Decomps'              , justify='right' )
        # fmt: on

        for m_id, m in enumerate(self.models):
            d = ModelList.get_model_summary(m)
            table.add_row(
                Text(f'{m_id}', style='bold'),
                frm_status_to_rich(d['frm_status']),
                Text(f'{d["is_frozen"]}', style='bold'),
                Text(f'{d["needs_frm"]}', style='bold'),
                Text(f'{d["frm_reasons"]}', style='bold'),
                Text(f'{d["rel_path"]}', style='bold'),
                Text(f'{d["src_last_ched"]}', style='bold'),
                Text(f'{d["user_iml_code"]}', style='bold'),
                Text(f'{d["user_iml_last_ched"]}', style='bold'),
                Text(f'{d["src_code_embds"]}', style='bold'),
                Text(f'{d["iml_code_embds"]}', style='bold'),
                Text(f'{d["context"]}', style='bold'),
                Text(f'{d["task_id"]}', style='bold'),
                Text(f'{d["deps"]}', style='bold'),
                Text(f'{d["num_opaques"]}', style='bold'),
                Text(f'{d["num_vgs"]}', style='bold'),
                Text(f'{d["num_decomps"]}', style='bold'),
            )

        return table



#
#   Imandra Inc.
#
#   cmd_models.py
#

from logging import getLogger
from typing import Annotated

import typer
from rich import print as rprint

from .utils import CLServerClient

log = getLogger(__name__)

app = typer.Typer()


@app.command('cmd')
def run_metamodel_cmd(
    cmd: Annotated[str, typer.Option(help='Command to execute')] = '',
    addr: Annotated[str, typer.Option(help='Server address')] = 'http://127.0.0.1:8000',
):
    """
    Run search command
    """

    cl_client = CLServerClient(addr)

    response = cl_client.get('search', {'query': cmd})

    if len(response.json()) == 0:
        print('No results found!')
    else:
        for res in response.json():
            rprint(res)



#
#   Imandra Inc.
#
#   sketch.py
#

import uuid

from imandra.u.agents.code_logician.base import FormalizationStatus
from imandra.u.agents.code_logician.base.region_decomp import RegionDecomp
from imandra.u.agents.code_logician.base.vg import VG
from pydantic import BaseModel, Field
from rich.panel import Panel
from rich.pretty import Pretty

from codelogician.util import find, maybe, maybe_else

# from imandra.u.agents.code_logician.base.imandrax import DecomposeRes, EvalRes, PO_Res
from ..tools.iml_utils import add_definition, remove_definition, update_definition
from .model import Model
from .sketch_task import (
    SketchChange,
    SketchChangeResult,
    SketchChangeTask,
    SketchChgDeleteDef,
    SketchChgInsertDef,
    SketchChgModifyDef,
    SketchChgSetModel,
)


class VGChange(BaseModel):
    """
    Helps us track VG changes over different states
    """

    is_new: bool  # Is this a new VG that was just added?
    is_removed: bool  # Has this VG been removed?
    res_changed: tuple[str, str] | None = None  # old vs new result type
    vg: VG  # Actual VG info


class ChangeSummary(BaseModel):
    """
    Summary of the formalization state changes since the last state where we had an admissable model
    """

    vg_changes: list[VGChange]

    def stats(self) -> dict[str, int]:
        """
        Return VG change statistics
        """

        def status_to_int(s: str) -> int:
            if s == 'unknown':
                return 1
            elif s == 'error' or s == 'err':
                return 2
            elif s == 'refuted':
                return 3
            elif s == 'verified_upto':
                return 4
            elif s == 'proved':
                return 5
            else:
                raise Exception(f'Unknown VG status: >{s}<')

        improved, deteriorated, proved, refuted = 0, 0, 0, 0
        for vg_change in self.vg_changes:
            if vg_change.is_removed or vg_change.res_changed is None:
                continue

            if vg_change.res_changed[1] in ['proved', 'verified_upto']:
                proved += 1
            if vg_change.res_changed[1] in ['refuted']:
                refuted += 1

            if not vg_change.is_new:
                if status_to_int(vg_change.res_changed[0]) > status_to_int(
                    vg_change.res_changed[1]
                ):
                    deteriorated += 1
                elif status_to_int(vg_change.res_changed[0]) > status_to_int(
                    vg_change.res_changed[1]
                ):
                    improved += 1

        return {
            'added': sum(1 if vgc.is_new else 0 for vgc in self.vg_changes),
            'removed': sum(1 if vgc.is_removed else 0 for vgc in self.vg_changes),
            'improved': improved,
            'deteriorated': deteriorated,
            'proved': proved,
            'refuted': refuted,
        }


def calc_vg_difference(old_vgs: list[VG], new_vgs: list[VG]) -> list[VGChange]:
    """
    Compute difference between two lists of VGs
    """
    vg_diff: list[VGChange] = []

    def find_a_match(new_v: VG, other_vgs: list[VG]) -> VG | None:
        def aux(new_data):
            def eq_preds(data):
                return data.predicate == new_data.predicate

            return find(lambda x: maybe_else(False, eq_preds, x.data), other_vgs)

        return maybe(aux, new_v.data)

    for new_vg in new_vgs:
        if new_vg.res is None:
            continue

        existing_vg = find_a_match(new_vg, old_vgs)

        if existing_vg is None:
            vg_diff.append(
                VGChange(
                    is_new=True,
                    is_removed=False,
                    res_changed=('', new_vg.res.res_type),
                    vg=new_vg,
                )
            )
        else:
            # This is an existing VG, let's see if its result type has changed at all...
            if (
                existing_vg.res is not None
                and new_vg.res.res_type != existing_vg.res.res_type
            ):
                vg_diff.append(
                    VGChange(
                        is_new=False,
                        is_removed=False,
                        res_changed=(existing_vg.res.res_type, new_vg.res.res_type),
                        vg=new_vg,
                    )
                )
            # ok, nothing changed
            else:
                vg_diff.append(
                    VGChange(
                        is_new=False, is_removed=False, res_changed=None, vg=new_vg
                    )
                )

    for old_vg in old_vgs:
        if old_vg.res is None:
            continue

        if find_a_match(old_vg, new_vgs) is None:
            vg_diff.append(
                VGChange(
                    is_new=False,
                    is_removed=True,
                    res_changed=(old_vg.res.res_type, ''),
                    vg=old_vg,
                )
            )

    return vg_diff


class SketchState(BaseModel):
    """
    SketchState contains the current version of
    """

    state_id: int  # State ID
    change: SketchChange  # Change that caused this state
    iml_code: str  # Actual IML code
    status: FormalizationStatus  # Formalization status
    error: str | None = None  # Error feedback from ImandraX

    vgs: list[VG]  # List of current verification goals
    old_vgs: list[VG]  # List of VGs in the last (good) state with admissable model
    decomps: list[RegionDecomp] = []  # Dictionary of decomposition requests/results

    vg_changes: list[VGChange] = Field(
        default_factory=lambda data: calc_vg_difference(data['vgs'], data['old_vgs'])
        if 'vgs' in data and 'old_vgs' in data
        else []
    )  # Changes in VGs

    # Summary of VG changes
    summary: ChangeSummary = Field(
        default_factory=lambda data: ChangeSummary(vg_changes=data['vg_changes'])
    )


class Sketch(BaseModel):
    """
    Sketch represent versions of a model
    """

    sketch_id: str = Field(default_factory=lambda: str(uuid.uuid4())[:8])
    processing: bool = False  # TODO do we need this for the UI?
    anchor_model_path: str  # relative path of the model that was used as the anchor
    init_iml_model: str = ''  # the initial model
    sketch_states: list[SketchState] = []  # here we keep track of changes in the state
    base_models: dict[
        str, Model
    ]  # Dict (by relative path) of base models that went into this

    def get_state_by_idx(self, idx: int) -> SketchState:
        """
        Return SketchState for specific state ID
        """

        if not (0 <= idx and idx < len(self.sketch_states)):
            raise ValueError(f'Out of bounds state id: {idx}')

        return self.sketch_states[idx]

    def rollback(self, target_state_id: int | None = None):
        """
        rollback states
        """

        if target_state_id is None:
            # We should just be able to
            if len(self.sketch_states) in [0, 1]:
                raise Exception(
                    f'Not enough states[={len(self.sketch_states)}] to rollback!'
                )
            self.sketch_states = self.sketch_states[1:]

        else:
            if not (0 <= target_state_id and target_state_id < len(self.sketch_states)):
                raise Exception(
                    f'Invalid target state id[={target_state_id}] specified!'
                )

            self.sketch_states = self.sketch_states[: (target_state_id + 1)]

    def state_ids(self) -> list[int]:
        """
        Return list of SketchState IDs
        """
        if self.sketch_states:
            return list(range(len(self.sketch_states)))
        return []

    def latest_state(self) -> SketchState | None:
        """
        Return the latest sketch state if it's available
        """
        return self.sketch_states[0] if len(self.sketch_states) else None

    def last_good_state(self) -> SketchState | None:
        """
        Return last sketch state that has an admissable model
        """
        for i in range(len(self.sketch_states) - 1, -1, -1):
            if self.sketch_states[i].status == FormalizationStatus.TRANSPARENT:
                return self.sketch_states[i]

        return None

    def gen_init_task(self) -> SketchChangeTask:
        """
        This is the task that needs to be run to populate the initial state
        """
        return SketchChangeTask(
            sketch_id=self.sketch_id,
            iml_code=self.init_iml_model,
            change=SketchChgSetModel(new_iml_code=self.init_iml_model),
        )

    def get_task(self, change: SketchChange) -> SketchChangeTask:
        """
        Return a SketchChangeTask that can be sent to the worker to be analyzed by ImandraX
        """

        # Writing it this way so it's easier for pyright to pick up
        state = self.latest_state()

        if state is None:
            raise Exception('Attempting to create a task but we have no initial state!')

        new_iml_model = self.process_change(state.iml_code, change)

        # Now we have a sketch change task!
        return SketchChangeTask(
            sketch_id=self.sketch_id, change=change, iml_code=new_iml_model
        )

    def apply_change_result(self, change_result: SketchChangeResult) -> ChangeSummary:
        """
        Apply the result of running change task to the state and get the ChangeSummary as the result
        """
        if len(self.sketch_states) == 0 or self.last_good_state() is None:
            state_id = 0
            old_vgs = []
            self.init_iml_model = change_result.task.iml_code
        else:
            state_id = len(self.sketch_states) + 1
            old_vgs = self.last_good_state().vgs  # pyright: ignore

        new_state = SketchState(
            state_id=state_id,
            change=change_result.task.change,
            iml_code=change_result.task.iml_code,
            status=FormalizationStatus.TRANSPARENT
            if change_result.success
            else FormalizationStatus.INADMISSIBLE,
            error=change_result.error,
            vgs=change_result.vgs,
            old_vgs=old_vgs,
            decomps=change_result.decomps,
        )

        self.sketch_states.insert(0, new_state)

        return new_state.summary

    def get_model_changes(self) -> dict[str, str]:
        """
        Generate changes to the underlying models
        """

        return {}

    def process_change(self, iml_code: str, change: SketchChange) -> str:
        """
        Process the change command for the specified code and return its updated version
        """

        if isinstance(change, SketchChgSetModel):
            updated_iml_code = change.new_iml_code

        elif isinstance(change, SketchChgInsertDef):
            try:
                updated_iml_code = add_definition(
                    source=iml_code, new_def=change.new_def_code
                )
            except Exception as e:
                raise Exception(f'Failed to insert: {e}')

        elif isinstance(change, SketchChgModifyDef):
            try:
                if isinstance(change.def_name_or_vg_idx, int):
                    updated_iml_code = update_definition(
                        source=iml_code,
                        name='',
                        new_def=change.new_def_body,
                        verify_index=change.def_name_or_vg_idx,
                    )
                else:
                    updated_iml_code = update_definition(
                        source=iml_code,
                        name=change.def_name_or_vg_idx,
                        new_def=change.new_def_body,
                    )
            except Exception as e:
                raise Exception(f'Failed to update definition: {e}')

        elif isinstance(change, SketchChgDeleteDef):
            try:
                if isinstance(change.def_name_or_vg_idx, int):
                    updated_iml_code = remove_definition(
                        source=iml_code, name='', verify_index=change.def_name_or_vg_idx
                    )
                else:
                    updated_iml_code = remove_definition(
                        source=iml_code, name=change.def_name_or_vg_idx
                    )
            except Exception as e:
                raise Exception(f'Failed to delete definition: {e}')

        else:
            raise Exception(f'Unsupported change event: {change}')

        return updated_iml_code

    def __repr__(self):
        return f'Sketch [ID = {self.sketch_id}; Anchor model path = {self.anchor_model_path}]'


class SketchContainer(BaseModel):
    """
    Container for helping us keep track of the sketches
    """

    sketches: dict[str, Sketch] = {}

    def add(self, new_sketch: Sketch) -> None:
        """
        Add spetch to the container
        """
        if new_sketch.sketch_id in self.sketches:
            raise ValueError(
                f'Sketch with specified ID is already present: {new_sketch.sketch_id}'
            )

        self.sketches[new_sketch.sketch_id] = new_sketch

    def from_id(self, sketch_id: str) -> Sketch | None:
        """
        Return sketch from sketch ID, None if it's missing
        """
        return self.sketches.get(sketch_id)

    def ids(self) -> list[str]:
        """
        List of available sketch IDs
        """
        return list(self.sketches.keys())

    def list_sketches(self) -> list[dict[str, str]]:
        """
        List available sketches
        """
        return [
            {
                'sketch_id': sketch_id,
                'anchor_model': self.sketches[sketch_id].anchor_model_path,
            }
            for sketch_id in self.sketches
        ]

    def toJSON(self):
        """
        Return JSON dict
        """
        return self.model_dump_json()

    @staticmethod
    def fromJSON(j: str | dict):
        """
        Create a SketchesContainer object from json dict
        """
        if isinstance(j, str):
            return SketchContainer.model_validate_json(j)
        else:
            return SketchContainer.model_validate(j)

    def __rich__(self):
        """Return the list of rich representations of the sketches we hold"""

        pretty = Pretty(self.toJSON(), indent_guides=True)
        return Panel(pretty, title='SketchContainer', border_style='green')



#
#   Imandra Inc.
#
#   model_task.py
#

import datetime
import uuid
from enum import StrEnum

from imandra.u.agents.code_logician.base import FormalizationDependency
from imandra.u.agents.code_logician.command import (
    AgentFormalizerCommand,
    Command,
    EditStateElementCommand,
    EmbedCommand,
    GenFormalizationDataCommand,
    GenModelCommand,
    GenRegionDecompsCommand,
    GenVgsCommand,
    InitStateCommand,
    InjectFormalizationContextCommand,
    SetModelCommand,
)
from imandra.u.agents.code_logician.graph import GraphState
from pydantic import BaseModel, Field


class ModelTaskMode(StrEnum):
    """
    Strategy mode
    """

    AGENT = 'Agent'
    MANUAL = 'Manual'


class FormalizationNeed(StrEnum):
    """
    Reasons to perform formalization
    """

    # fmt: off
    NO_AGENT_STATE = 'No_agent_state' # This is the first time we're formalizing the model
    SRC_CODE_CHANGED = 'Src_code_changed' # Source code has changed
    IML_CODE_CHANGED = 'IML_code_changed' # User edited the IML code, so we need to resend it to CL
    CONTEXT_ADDED = 'Cntxt_added' # Human feedback has been provided
    DEPS_CHANGED = 'Deps_changed' # Dependencies (at least one) have changed
    # fmt: on


class ModelTask(BaseModel):
    """
    Model task contains a single task for Code Logician to execute
    """

    # fmt: off
    rel_path        : str # path of the source file (relative to the source code directory)
    src_code        : str = '' # source code
    context         : str | None = '' # IML source code for dependent models
    dependencies    : list[FormalizationDependency] = [] # IML source code for dependent models
    graph_state     : GraphState | None = None # previous graph state, if available
    language        : str = 'Python' # programming language
    mode            : ModelTaskMode = ModelTaskMode.AGENT # mode used with CodeLogician (either 'simple' or 'agent')
    gen_vgs         : bool = False # Should we generate verification goals?
    gen_decomps     : bool = False # Should we generate decompositions?
    gen_embeddings  : bool = True # Should we generate embeddings?
    # fmt: on

    # This is used to submit user-specified commands for specific models
    # if this is set, then we disregard everything else and just return these
    # along with the existing state
    specified_commands: list[Command] | None = None

    # Each task has a unique ID that we'll then use to assign back the result
    # of CL work for this task - if the model's task ID doesn't match the result
    # then we discard it - this way we always keep the most recent result
    task_id: str = Field(
        default_factory=lambda: str(uuid.uuid4())
    )  # assigned task ID, new one created if not provided

    # Same with the timestamp, if it's provided (mostly during de/serialization,
    # then let's use it otherwise, let's create a new one)
    created_at: datetime.datetime = Field(
        default_factory=datetime.datetime.now
    )  # Each instance gets a new timestamp

    def commands(self):
        """
        Return a list of CodeLogician agent commands for this task
        """

        def yield_commands():
            yield InitStateCommand(src_code=self.src_code, src_lang=self.language)

            if self.context is not None:
                yield InjectFormalizationContextCommand(context=self.context)

            if self.dependencies:
                yield EditStateElementCommand(update={'dependency': self.dependencies})

            if self.mode == ModelTaskMode.MANUAL:
                yield GenFormalizationDataCommand()
                yield GenModelCommand()

            elif self.mode == ModelTaskMode.AGENT:
                yield AgentFormalizerCommand(
                    no_gen_model_hitl=True,
                    max_tries_wo_hitl=3,
                    max_tries=3,
                    no_check_formalization_hitl=True,
                    no_refactor=False,
                )
            else:
                raise Exception(f'Attempting unrecognized mode: {self.mode}')

            if self.gen_vgs:
                yield GenVgsCommand(description='')

            if self.gen_decomps:
                yield GenRegionDecompsCommand(function_name=None)
                # for i in self.tests:
                #     yield GenTestCasesCommand(decomp_idx=i)

            if self.gen_embeddings:
                yield EmbedCommand()

        return self.specified_commands or list(yield_commands())

    def toJSON(self):
        """
        Convert to a JSON
        """
        return self.model_dump_json()

    @staticmethod
    def fromJSON(j: dict | str):
        """
        fromJSON
        """
        if isinstance(j, str):
            return ModelTask.model_validate_json(j)
        else:
            return ModelTask.model_validate(j)

    def __repr__(self):
        """ """
        return f'Base ModelTask with ID = {self.task_id}; path={self.rel_path}'


class UserManualIMLEditTask(ModelTask, BaseModel):
    """
    This updates the graph state from a manual user IML edit
    """

    iml_code: str  # user provided IML code

    def commands(self):
        """
        We're just going to invoke the commands
        """

        if self.graph_state:
            commands = [SetModelCommand(model=self.iml_code)]
        else:
            commands: list[Command] = [
                InitStateCommand(src_code=self.src_code, src_lang=self.language),
                SetModelCommand(model=self.iml_code),
            ]

        return commands
